import enum
from importlib.metadata import entry_points

import numpy as np
from calngUtils import geom_utils
from karabo.bound import (
    BOOL_ELEMENT,
    KARABO_CLASSINFO,
    NODE_ELEMENT,
    OUTPUT_CHANNEL,
    OVERWRITE_ELEMENT,
    STRING_ELEMENT,
    TABLE_ELEMENT,
    VECTOR_BOOL_ELEMENT,
    Epochstamp,
    Hash,
    Schema,
    State,
    Timestamp,
    Trainstamp,
)
from TrainMatcher import TrainMatcher

from . import scenes
from ._version import version as deviceVersion
from .arbiter_kernels.base_kernel import KernelWarning
from .expr import BinaryExpression
from .utils import WarningContextSystem


class DeviceWarning(enum.Enum):
    PLAN = "plan"
    DECISION = "decision"


def output_schema():
    res = Schema()
    (
        NODE_ELEMENT(res)
        .key("data")
        .commit(),

        VECTOR_BOOL_ELEMENT(res)
        .key("data.dataFramePattern")
        .assignmentOptional()
        .defaultValue([])
        .commit(),
    )
    return res


kernels = [
    kernel_class.load()
    for kernel_class in entry_points().select().get("calng.arbiter_kernel", [])
]
kernel_choice = {kernel_class.__name__: kernel_class for kernel_class in kernels}


def selection_plan_schema(kernels):
    schema = Schema()
    (
        BOOL_ELEMENT(schema)
        .key("enable")
        .displayedName("Enable")
        .assignmentOptional()
        .defaultValue(False)
        .reconfigurable()
        .commit(),

        STRING_ELEMENT(schema)
        .key("name")
        .displayedName("Name")
        .description(
            "The name assigned to the result of running this kernel. This name can be "
            "used in (later) preselection expressions and in the final decision "
            "expression. Each row must have a unique name and names must be valid "
            "identifiers. Each selection step gets its configuration from a schema "
            "node corresponding to its name."
        )
        .assignmentOptional()
        .defaultValue("name")
        .reconfigurable()
        .commit(),

        STRING_ELEMENT(schema)
        .key("kernel")
        .displayedName("Kernel")
        .description(
            "Choose which kernel to run in this selection step. The special kernel "
            "'Assign' does nothing, but by using the preselection expression, it "
            "effectively assigns the result of the preselection to the given name, "
            "allowing you to name combined results from other kernels."
        )
        .options(",".join(kernel_choice))
        .assignmentOptional()
        .defaultValue("Assign")
        .reconfigurable()
        .commit(),

        STRING_ELEMENT(schema)
        .key("preselection")
        .displayedName("Preselection")
        .description(
            "If not empty, this is evaluated as a logical expression (see "
            "frameSelection.decision) involving previous selection steps. The "
            "resulting mask is then passed to the kernel running in this selection "
            "step. It is up to the kernel how the mask is used. The 'Assign' kernel "
            "will just return the mask it is given."
        )
        .assignmentOptional()
        .defaultValue("")
        .reconfigurable()
        .commit(),
    )
    return schema


class BaseFrameSelectionArbiter(TrainMatcher.TrainMatcher):
    @staticmethod
    def expectedParameters(expected):
        (
            STRING_ELEMENT(expected)
            .key("geometryDevice")
            .displayedName("Geometry device")
            .description(
                "The the device which will provide geometry."
            )
            .assignmentMandatory()
            .commit(),

            NODE_ELEMENT(expected)
            .key("frameSelection")
            .displayedName("Frame selection")
            .commit(),
        )

    def __init__(self, config):
        super().__init__(config)

    def initialization(self):
        super().initialization()
        self.warning_context = WarningContextSystem(
            self, on_success={"state": "PASSIVE"}
        )

        self._geometry = None

    def _guess_number_of_frames(self, sources):
        # TODO: robust frame deduction
        for (data, _) in sources.values():
            if data.has("image.cellId"):
                return data.get("image.cellId").size
        self.log.ERROR("Unable to figure out number of frames")


@KARABO_CLASSINFO("SimpleFrameSelectionArbiter", deviceVersion)
class SimpleFrameSelectionArbiter(BaseFrameSelectionArbiter):
    @staticmethod
    def expectedParameters(expected):
        (
            STRING_ELEMENT(expected)
            .key("frameSelection.kernelChoice")
            .options(",".join(kernel_choice))
            .assignmentOptional()
            .defaultValue("ManualFilter")
            .reconfigurable()
            .commit(),

            STRING_ELEMENT(expected)
            .key("frameSelection.kernelState")
            .setSpecialDisplayType("State")
            .readOnly()
            .initialValue("OFF")
            .commit(),
        )

    def initialization(self):
        super().initialization()

        self._update_schema(self["frameSelection.kernelChoice"])
        self._initialize_kernel(
            self["frameSelection.kernelChoice"], self["frameSelection.kernelParameters"]
        )
        geom_utils.subscribe_to_geometry_bound(
            self, self.get("geometryDevice"), on_geometry=self._on_geometry
        )

        if self["state"] != State.ERROR:
            self.start()  # Auto-start this type of matcher.

    def _update_schema(self, kernel_name):
        schema_update = Schema()
        (
            NODE_ELEMENT(schema_update)
            .key("frameSelection")
            .commit(),

            NODE_ELEMENT(schema_update)
            .key("frameSelection.kernelParameters")
            .commit(),
        )
        kernel_class = kernel_choice[kernel_name]
        kernel_class.extend_device_schema(
            schema_update, "frameSelection.kernelParameters"
        )
        output_schema_update = output_schema()
        kernel_class.extend_output_schema(output_schema_update, kernel_class.__name__)
        (
            OUTPUT_CHANNEL(schema_update)
            .key("output")
            .dataSchema(output_schema_update)
            .commit(),
        )
        self.updateSchema(schema_update)

    def _initialize_kernel(self, class_name, conf):
        kernel_class = kernel_choice[class_name]
        self.kernel = kernel_class(self, kernel_class.__name__, conf)

    def preReconfigure(self, conf):
        super().preReconfigure(conf)

        if conf.has("frameSelection.kernelChoice"):
            name = conf["frameSelection.kernelChoice"]
            self._update_schema(name)
            self._initialize_kernel(name, self["frameSelection.kernelParameters"])
        if conf.has("frameSelection.kernelParameters"):
            self.kernel.reconfigure(conf["frameSelection.kernelParameters"])

    def on_matched_data(self, train_id, sources):
        num_frames = self._guess_number_of_frames(sources)

        # run kernels
        out_hash = Hash()
        with self.warning_context(
            "frameSelection.kernelState", KernelWarning.PROCESSING
        ) as warn:
            try:
                decision = self.kernel.consider(
                    train_id,
                    sources,
                    num_frames,
                    np.ones(num_frames, dtype=bool),
                    out_hash,
                )
            except Exception as ex:
                warn(f"Kernel failed: {ex}")
                decision = np.ones(num_frames, dtype=bool)

        out_hash["data.dataFramePattern"] = list(map(bool, decision))

        # don't use TM's stored self.output as we have re-injected
        self.writeChannel(
            "output",
            out_hash,
            timestamp=Timestamp(Epochstamp(), Trainstamp(train_id)),
        )
        self.info["sent"] += 1
        self.info["trainId"] = train_id
        self.rate_out.update()

    def _on_geometry(self, geometry):
        self.kernel.on_new_geometry(geometry)


@KARABO_CLASSINFO("AdvancedFrameSelectionArbiter", deviceVersion)
class AdvancedFrameSelectionArbiter(BaseFrameSelectionArbiter):
    @staticmethod
    def expectedParameters(expected):
        (
            OVERWRITE_ELEMENT(expected)
            .key("availableScenes")
            .setNewDefaultValue(["overview", "trainMatcherScene"])
            .commit(),

            TABLE_ELEMENT(expected).key("frameSelection.plan")
            .displayedName("Selection Plan")
            .description("")
            .setColumns(selection_plan_schema(kernel_choice.keys()))
            .assignmentOptional()
            .defaultValue([])
            .reconfigurable()
            .commit(),

            STRING_ELEMENT(expected)
            .key("frameSelection.decision")
            .displayedName("Decision")
            .assignmentMandatory()
            .reconfigurable()
            .commit(),

            NODE_ELEMENT(expected)
            .key("frameSelection.selections")
            .displayedName("Selections")
            .commit(),
        )

    def __init__(self, config):
        super().__init__(config)
        self._selection_steps = {}  # name -> (kernel class, preselection expression)
        self._selection_kernels = {}  # name -> kernel instance
        self._decision_expr = BinaryExpression(None)

    def initialization(self):
        super().initialization()

        self._validate_plan_and_update_schema(self["frameSelection.plan"])
        self._initialize_kernels()
        self._configure_decision(self["frameSelection.decision"])
        geom_utils.subscribe_to_geometry_bound(
            self, self.get("geometryDevice"), on_geometry=self._on_geometry
        )

        if self["state"] != State.ERROR:
            self.start()  # Auto-start this type of matcher.

    def start(self):
        super().start()
        self.warning_context.on_success["state"] = "ACTIVE"

    def stop(self):
        super().stop()
        self.warning_context.on_success["state"] = "PASSIVE"

    def requestScene(self, params):
        scene_name = params.get("name", default="")
        if scene_name == "trainMatcherScene":
            params["name"] = "scene"
            return super().requestScene(params)
        if scene_name != "overview":
            self.log.INFO(
                f"Unexpected scene name {scene_name} requested, "
                "will just return overview"
            )
        payload = Hash("name", scene_name, "success", True)
        payload["data"] = scenes.complex_arbiter_overview(
            device_id=self.getInstanceId(),
            schema=self.getFullSchema(),
        )
        self.reply(
            Hash(
                "type",
                "deviceScene",
                "origin",
                self.getInstanceId(),
                "payload",
                payload,
            )
        )

    def on_matched_data(self, train_id, sources):
        num_frames = self._guess_number_of_frames(sources)

        # run kernels
        out_hash = Hash()
        selections = self._process_selection_steps(sources, train_id, num_frames, out_hash)

        # final selection decision
        decision = self._decision_expr.eval(selections, np.ones(num_frames, bool))
        out_hash["data.dataFramePattern"] = list(map(bool, decision))

        # don't use TM's stored self.output as we have re-injected
        self.writeChannel(
            "output",
            out_hash,
            timestamp=Timestamp(Epochstamp(), Trainstamp(train_id)),
        )
        self.info["sent"] += 1
        self.info["trainId"] = train_id
        self.rate_out.update()

    def _process_selection_steps(self, sources, train_id, num_frames, out_hash):
        res = {}
        for selection_name, kernel in self._selection_kernels.items():
            prefix = f"frameSelection.selections.{selection_name}"
            with self.warning_context(
                f"{prefix}.state",
                KernelWarning.PRESELECTION,
            ) as warn:
                default_mask = np.ones(num_frames, dtype=bool)
                _, selection_expr = self._selection_steps[selection_name]
                try:
                    mask = selection_expr.eval(
                        res, default_mask
                    )
                except Exception as ex:
                    warn(f"Evaluating preselection for {prefix} failed: {ex}")
                    mask = default_mask
            with self.warning_context(
                f"{prefix}.state",
                KernelWarning.PROCESSING,
            ) as warn:
                try:
                    res[selection_name] = kernel.consider(
                        train_id, sources, num_frames, mask, out_hash
                    )
                except Exception as ex:
                    warn(f"Kernel for {prefix} failed: {ex}")
                    res[selection_name] = np.ones(num_frames, dtype=bool)
        return res

    def _configure_decision(self, decision_string):
        with self.warning_context("state", DeviceWarning.DECISION) as warn:
            try:
                self._decision_expr = BinaryExpression.parse(
                    decision_string, self._selection_steps.keys()
                )
            except Exception as ex:
                # TODO: maybe re-raise to reject reconfiguration
                warn(f"Failure to parse decision string '{decision_string}': {ex}")

    def _validate_plan_and_update_schema(self, plan):
        new_selection_steps = {}

        # TODO: don't update this many levels, just updaet frameSelection.selections
        schema_update = Schema()
        (
            NODE_ELEMENT(schema_update)
            .key("frameSelection")
            .displayedName("Frame selection")
            .commit(),

            NODE_ELEMENT(schema_update)
            .key("frameSelection.selections")
            .displayedName("Selections")
            .commit(),
        )
        output_schema_update = output_schema()
        with self.warning_context("state", DeviceWarning.PLAN) as warn:
            for selection in plan:
                if not selection["enable"]:
                    continue

                selection_name = selection["name"].strip()
                if not selection_name.isidentifier():
                    warn(f"Invalid selection name '{selection_name}'")
                    continue
                elif selection_name in new_selection_steps:
                    warn(f"Duplicate selection name '{selection_name}'")
                    continue

                try:
                    preselection = BinaryExpression.parse(
                        selection["preselection"],
                        new_selection_steps,  # just needs to check names
                    )
                except Exception as ex:
                    warn(
                        "Failure to parse (or undefined variable used in) "
                        f"preselection for {selection_name}: {ex}"
                    )
                    continue

                kernel_class = kernel_choice[selection["kernel"]]

                kernel_prefix = f"frameSelection.selections.{selection_name}"
                (
                    NODE_ELEMENT(schema_update)
                    .key(kernel_prefix)
                    .commit(),

                    STRING_ELEMENT(schema_update)
                    .key(f"{kernel_prefix}.kernel")
                    .readOnly()
                    .initialValue(kernel_class.__name__)
                    .commit(),

                    STRING_ELEMENT(schema_update)
                    .key(f"{kernel_prefix}.state")
                    .setSpecialDisplayType("State")
                    .readOnly()
                    .initialValue("OFF")
                    .commit(),
                )
                kernel_class.extend_device_schema(
                    schema_update, kernel_prefix
                )
                kernel_class.extend_output_schema(output_schema_update, selection_name)
                new_selection_steps[selection_name] = (kernel_class, preselection)

        (
            OUTPUT_CHANNEL(schema_update)
            .key("output")
            .dataSchema(output_schema_update)
            .commit(),
        )
        self._selection_steps = new_selection_steps
        self.updateSchema(schema_update)

    def _initialize_kernels(self):
        # instantiate kernels for selections
        self._selection_kernels.clear()
        for selection_name, (
            kernel_class,
            preselection,
        ) in self._selection_steps.items():
            prefix = f"frameSelection.selections.{selection_name}"
            with self.warning_context(f"{prefix}.state", KernelWarning.INIT) as warn:
                try:
                    self._selection_kernels[selection_name] = kernel_class(
                        self, selection_name, self.get(prefix)
                    )
                except Exception as ex:
                    warn(f"Failed to instantiate kernel for {selection_name}: {ex}")

    def preReconfigure(self, conf):
        super().preReconfigure(conf)

        if conf.has("frameSelection.plan"):
            self._validate_plan_and_update_schema(conf["frameSelection.plan"])
            self._initialize_kernels()

        if conf.has("frameSelection.decision"):
            self._configure_decision(conf["frameSelection.decision"])

        if conf.has("frameSelection.selections"):
            for name, kernel in self._selection_kernels.items():
                kernel_prefix = f"frameSelection.selections.{name}"
                if conf.has(kernel_prefix):
                    kernel.reconfigure(conf[kernel_prefix])

    def _on_geometry(self, geometry):
        for kernel in self._selection_kernels.values():
            kernel.on_new_geometry(geometry)
